"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolanaContracts = void 0;
const tokenBridge_1 = require("./utils/tokenBridge");
const wormhole_1 = require("./utils/wormhole");
const nftBridge_1 = require("./utils/nftBridge");
/**
 * @category Solana
 */
class SolanaContracts {
    _contracts;
    constructor(conf) {
        this._contracts = new Map();
        Object.entries(conf).forEach(([c, cfg]) => {
            this._contracts.set(c, cfg.contracts);
        });
    }
    getContracts(chain) {
        return this._contracts.get(chain);
    }
    mustGetContracts(chain) {
        const contracts = this._contracts.get(chain);
        if (!contracts)
            throw new Error(`no Solana contracts found for ${chain}`);
        return contracts;
    }
    /**
     * Returns core wormhole contract for the chain
     *
     * @returns An interface for the core contract, undefined if not found
     */
    getCore(chain, connection) {
        const contracts = this.mustGetContracts(chain);
        if (!contracts.coreBridge)
            return;
        return (0, wormhole_1.createReadOnlyWormholeProgramInterface)(contracts.coreBridge, connection);
    }
    /**
     * Returns core wormhole contract for the chain
     *
     * @returns An interface for the core contract, errors if not found
     */
    mustGetCore(chain, connection) {
        const core = this.getCore(chain, connection);
        if (!core)
            throw new Error(`Core contract for domain ${chain} not found`);
        return core;
    }
    /**
     * Returns wormhole bridge contract for the chain
     *
     * @returns An interface for the bridge contract, undefined if not found
     */
    getTokenBridge(chain, connection) {
        const contracts = this.mustGetContracts(chain);
        if (!contracts.tokenBridge)
            return;
        return (0, tokenBridge_1.createReadOnlyTokenBridgeProgramInterface)(contracts.tokenBridge, connection);
    }
    /**
     * Returns wormhole bridge contract for the chain
     *
     * @returns An interface for the bridge contract, errors if not found
     */
    mustGetTokenBridge(chain, connection) {
        const bridge = this.getTokenBridge(chain, connection);
        if (!bridge)
            throw new Error(`Bridge contract for domain ${chain} not found`);
        return bridge;
    }
    /**
     * Returns wormhole NFT bridge contract for the chain
     *
     * @returns An interface for the NFT bridge contract, undefined if not found
     */
    getNftBridge(chain, connection) {
        const contracts = this.mustGetContracts(chain);
        if (!contracts.nftBridge)
            return;
        return (0, nftBridge_1.createReadOnlyNftBridgeProgramInterface)(contracts.nftBridge, connection);
    }
    /**
     * Returns wormhole NFT bridge contract for the chain
     *
     * @returns An interface for the NFT bridge contract, errors if not found
     */
    mustGetNftBridge(chain, connection) {
        const nftBridge = this.getNftBridge(chain, connection);
        if (!nftBridge)
            throw new Error(`NFT Bridge contract for domain ${chain} not found`);
        return nftBridge;
    }
}
exports.SolanaContracts = SolanaContracts;
//# sourceMappingURL=contracts.js.map