"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolanaPlatform = void 0;
const web3_js_1 = require("@solana/web3.js");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const contracts_1 = require("./contracts");
const chain_1 = require("./chain");
const tokenBridge_1 = require("./protocols/tokenBridge");
const SOLANA_SEQ_LOG = 'Program log: Sequence: ';
/**
 * @category Solana
 */
class SolanaPlatform {
    static _platform = 'Solana';
    platform = SolanaPlatform._platform;
    conf;
    contracts;
    constructor(conf) {
        this.conf = conf;
        this.contracts = new contracts_1.SolanaContracts(conf);
    }
    getRpc(chain, commitment = 'confirmed') {
        const rpcAddress = this.conf[chain].rpc;
        return new web3_js_1.Connection(rpcAddress, commitment);
    }
    getChain(chain) {
        return new chain_1.SolanaChain(this, chain);
    }
    async getDecimals(chain, rpc, token) {
        if (token === 'native')
            return BigInt(this.conf[chain]?.nativeTokenDecimals);
        let mint = await rpc.getParsedAccountInfo(new web3_js_1.PublicKey(token.address.unwrap()));
        if (!mint)
            throw new Error('could not fetch token details');
        const { decimals } = mint.value.data.parsed.info;
        return decimals;
    }
    async getBalance(chain, rpc, walletAddress, token) {
        if (token === 'native')
            return BigInt(await rpc.getBalance(new web3_js_1.PublicKey(walletAddress)));
        if (token.chain !== chain) {
            const tb = await this.getTokenBridge(rpc);
            token = { chain: chain, address: await tb.getWrappedAsset(token) };
        }
        const splToken = await rpc.getTokenAccountsByOwner(new web3_js_1.PublicKey(walletAddress), { mint: new web3_js_1.PublicKey(token.address.toUint8Array()) });
        if (!splToken.value[0])
            return null;
        const balance = await rpc.getTokenAccountBalance(splToken.value[0].pubkey);
        return BigInt(balance.value.amount);
    }
    async sendWait(rpc, stxns) {
        const txhashes = [];
        // TODO: concurrent?
        let lastTxHash;
        for (const stxn of stxns) {
            const txHash = await rpc.sendRawTransaction(stxn);
            // TODO: throw error?
            if (!txHash)
                continue;
            lastTxHash = txHash;
            txhashes.push(txHash);
        }
        // TODO: allow passing commitment level in? this method is also deprecated...
        await rpc.confirmTransaction(lastTxHash, 'finalized');
        return txhashes;
    }
    async getWormholeCore(rpc) {
        throw new Error('Not Supported');
        //return SolanaWormholeCore.fromProvider(rpc, this.contracts);
    }
    async getTokenBridge(rpc) {
        return tokenBridge_1.SolanaTokenBridge.fromProvider(rpc, this.contracts);
    }
    async getAutomaticTokenBridge(rpc) {
        throw new Error('Not Supported');
    }
    async getCircleBridge(rpc) {
        throw new Error('Not Supported');
    }
    async getAutomaticCircleBridge(rpc) {
        throw new Error('Not Supported');
    }
    parseAddress(chain, address) {
        return (0, connect_sdk_1.toNative)(chain, address);
    }
    async parseTransaction(chain, rpc, tx) {
        const contracts = this.contracts.mustGetContracts(chain);
        if (!contracts.coreBridge)
            throw new Error('contracts not found');
        const response = await rpc.getTransaction(tx);
        if (!response || !response.meta?.innerInstructions[0].instructions)
            throw new Error('transaction not found');
        const instructions = response.meta?.innerInstructions[0].instructions;
        const accounts = response.transaction.message.accountKeys;
        // find the instruction where the programId equals the Wormhole ProgramId and the emitter equals the Token Bridge
        const bridgeInstructions = instructions.filter((i) => {
            const programId = accounts[i.programIdIndex].toString();
            const wormholeCore = contracts.coreBridge;
            return programId === wormholeCore;
        });
        if (bridgeInstructions.length === 0)
            throw new Error('no bridge messages found');
        // TODO: unsure about the single bridge instruction and the [2] index, will this always be the case?
        const [logmsg] = bridgeInstructions;
        const emitterAcct = accounts[logmsg.accounts[2]];
        const emitter = this.parseAddress(chain, emitterAcct.toString());
        const sequence = response.meta?.logMessages
            ?.filter((msg) => msg.startsWith(SOLANA_SEQ_LOG))?.[0]
            ?.replace(SOLANA_SEQ_LOG, '');
        if (!sequence) {
            throw new Error('sequence not found');
        }
        return [
            {
                chain,
                emitter: emitter.toUniversalAddress(),
                sequence: BigInt(sequence),
            },
        ];
    }
}
exports.SolanaPlatform = SolanaPlatform;
//# sourceMappingURL=platform.js.map